//LPT-Server

//Dmon zur Steuerung der Parallelport-LED's

#include <stdio.h>
#include <string.h>
#include <asm/io.h>
#include <asm/types.h>
#include <asm/stat.h>
//#include <unistd.h>
#include <fcntl.h>

//#define _DEBUG_

#define D_LPT1_DATA   0x378
#define D_FIFO_FILE   "/var/tmp/lptprgf"
#define S_FEHLER_IO1  "Fehler: Kann nicht auf Schnittstelle zugreifen!\nBist du nicht root?\n"
#define S_FEHLER_IO2  "Fehler: Kann FIFO nicht anlegen: "
#define S_FEHLER_IO3  "Fehler: Kann FIFO nicht zum Lesen oeffnen: "

#define I_DATEN_NR    10
#define S_NULL        "NULL"
#define I_NULL        0
#define S_BOOT        "BOOT"
#define I_BOOT        1
#define S_NET         "NET"
#define I_NET         2
#define S_READY       "READY"
#define I_READY       4
#define S_PREPARE     "PREPARE"
#define I_PREPARE     8
#define S_USAGE1      "USAGE1"
#define I_USAGE1      16
#define S_USAGE2      "USAGE2"
#define I_USAGE2      32
#define S_AKTIV       "AKTIV"
#define I_AKTIV       64
#define S_ONLINE      "ONLINE"
#define I_ONLINE      128
#define S_ALL         "ALL"
#define I_ALL         255

char S_DATEN[I_DATEN_NR][8] = { S_NULL , S_BOOT, S_NET, S_READY, S_PREPARE, S_USAGE1, S_USAGE2, S_AKTIV, S_ONLINE, S_ALL };
int I_DATEN[I_DATEN_NR] = {I_NULL, I_BOOT, I_NET, I_READY, I_PREPARE, I_USAGE1, I_USAGE2, I_AKTIV, I_ONLINE, I_ALL };

#define I_BEFEHLE_NR  7
#define S_CTRL_NULL   "NULL"
#define I_CTRL_NULL   0
#define S_CTRL_QUIT   "QUIT"
#define I_CTRL_QUIT   1
#define S_CTRL_SET    "SET"
#define I_CTRL_SET    2
#define S_CTRL_ADD    "ADD"
#define I_CTRL_ADD    4
#define S_CTRL_DEL    "DEL"
#define I_CTRL_DEL    8
#define S_CTRL_ABLINK "ABLINK"
#define I_CTRL_ABLINK 16
#define S_CTRL_DBLINK "DBLINK"
#define I_CTRL_DBLINK 32

char S_BEFEHLE[I_BEFEHLE_NR][7] = { S_CTRL_NULL, S_CTRL_QUIT, S_CTRL_SET, S_CTRL_ADD, S_CTRL_DEL, S_CTRL_ABLINK, S_CTRL_DBLINK };
int I_BEFEHLE[I_BEFEHLE_NR] = { I_CTRL_NULL, I_CTRL_QUIT, I_CTRL_SET, I_CTRL_ADD, I_CTRL_DEL, I_CTRL_ABLINK, I_CTRL_DBLINK };

int delay_time = 1; //Wartezeit

int I_port = D_LPT1_DATA; //Datenport IO-Adresse
int r_fifo; //FIFO-Handle
int befehl = I_CTRL_NULL, daten = I_NULL;
int running = 1;
unsigned char dauer_an = 0, blink_maske = 0, oddBlink = 0, led_bits = 0;

inline void setzeSchnittstellenBits(); //Schreibt den aktuellen Wert von led_bits
inline void blinken(); //ndert den led_bits Wert um die Blinkmaske
void oeffneFIFO(); //erzeugt und ffnet FIFO
inline void holeNachrichten(); //lies den Inhalt der FIFO
void verarbeiteNachricht( char *nachricht ); //verarbeitet die empf. Nachricht

inline void debug(char *argv) {
    #ifdef _DEBUG_
    printf(argv);
    #endif
}

int kind() {
    int r = -1;
    debug("Kind\n");
#ifndef _DEBUG_
    if ( ioperm( I_port, 3, 1) != 0 ) {
	printf(S_FEHLER_IO1);
    } else {
#endif
	oeffneFIFO();
	while (running!=0) {
	    holeNachrichten();
	    blinken();
	    setzeSchnittstellenBits();
	    sleep( delay_time );
	}
	r = 0; // Return ok
#ifndef _DEBUG_
    }
#endif
    debug("Kind ist fertig\n");
    return r;
}

void verarbeiteNachricht( char *nachricht ) {
    int c1 = 0, c2=0, ti;
    debug("N>");
    debug(nachricht);
    for (c1=0; c1 < I_BEFEHLE_NR; c1++) {
	if (strcmp(nachricht, S_BEFEHLE[c1]) == 0) {
	    befehl = I_BEFEHLE[c1];
	}
    }
    for (c2=0; c2 < I_DATEN_NR; c2++) {
	if (strcmp(nachricht, S_DATEN[c2]) == 0) {
	    daten = I_DATEN[c2];
	}	
    }
    ti = atoi(nachricht);
    if (ti > 0) { daten = ti;}
    if (befehl == I_CTRL_QUIT) {running = 0;}
    if (befehl != I_CTRL_NULL && daten != I_NULL) {
	switch(befehl) {
	    case I_CTRL_SET: 
		debug("Kontrollsequenz Bits setzen");
		dauer_an = daten;
		break;
	    case I_CTRL_ADD:
		debug("Kontrollsequenz Bits addieren");
		dauer_an = dauer_an | daten;
		break;
	    case I_CTRL_DEL:
		debug("Kontrollsequenz Bits lschen");
		dauer_an = dauer_an & ( 255 - daten );
		break;
	    case I_CTRL_ABLINK:
		debug("Kontrollsequenz Bits blinken an");		
		blink_maske = blink_maske | daten;
		break;
	    case I_CTRL_DBLINK:
		debug("Kontrollsequenz Bits blinken aus");
		blink_maske = blink_maske & ( 255 - daten);
		break;
	}
	debug("\n");
	befehl = I_CTRL_NULL;
	daten = I_NULL;
    }
    if (daten != I_NULL) {
	befehl = I_CTRL_NULL;
	daten = I_NULL;
    }
}


void holeNachrichten() {
#define I_BUFSIZE     256
    char buf[I_BUFSIZE], i[2];
    int r, c = 0;
    strcpy(buf, ""); i[1] = 0;
    do {
	r = read(r_fifo, i, 1);
	if (r>0) { 
	    if (i[0] == ' ' ) {
		verarbeiteNachricht(buf);
		strcpy(buf, ""); c = 0;
	    }
	    if (i[0] > ' ' ) {
		strncat(buf, i, I_BUFSIZE-1); c++;
	    }
	}
    } while (r > 0 && c < I_BUFSIZE);
    if (strlen(buf) > 0) {verarbeiteNachricht(buf);}
    fflush(stdout);
}

void oeffneFIFO() {
    unlink(D_FIFO_FILE);
    debug("Erzeuge FIFO\n");
    if((mkfifo(D_FIFO_FILE, O_RDWR|0664)) == -1) {
	debug(S_FEHLER_IO2);
	debug(D_FIFO_FILE);
	debug("\n");
	exit(0);
    }

    debug("ffne FIFO\n");
    if (( r_fifo = open(D_FIFO_FILE, O_RDONLY)) < 0) {
	debug(S_FEHLER_IO3);
	debug(D_FIFO_FILE);
	debug("\n");
	exit(0); 
    }
    debug("FIFO offen\n");
}

void blinken() {
    oddBlink = !oddBlink;
    led_bits = dauer_an | (oddBlink ? 0 : blink_maske);
}

void setzeSchnittstellenBits() {
#ifndef _DEBUG_
    outb(led_bits, I_port);
#else
    printf("Port %i, Daten %i\n", I_port, led_bits);
#endif
}

int werdeDaemon() {
    int f, ret = 0;
    switch(f=fork())
    {
	case -1 : 
	    printf("Konnte kein Kindprozess erzeugen (fork).");
	    ret = -1;
	    break;
	case 0 : 
	    kind();
	    break;
	default : 
	    //Elternprozess, kann jetzt beendet werden
	    break;
    }
    return ( ret );
}

int main( int argc, char *argv[] )
{
    debug("Server startet...\n");
    //Prfe Kommandozeileargumente
    if (argc == 2) {
       debug("Es wurde ein Argument gefunden\n");
       I_port = atoi(argv[1]);
    }
    printf("Es wird Port %i verwendet.\n", I_port);
    //Prfe Rechte fr Zugriff auf Schnittstelle
#ifndef _DEBUG_
    if ( ioperm( I_port, 3, 1) != 0 ) {
       printf("Fehler: Kann auf Schnittstelle nicht zugreifen!\n(Bist du nicht root?)\n");
       return(-1);
       }
    //Werde Dmon
    return( werdeDaemon() );
#else
    kind();
    return(0); 
#endif
}
